OVERPASS = (("-u","--overpass-url"),{'type':str,'dest':'overpass_url','help':"define url for Overpass API to be URL",'metavar':'URL'})
NOGT = (("-g","--no-graph-tool"),{'action':'store_true','dest':'no_gt','default':False,'help':"do not perform graph_tool tests"})
CONFIG = [OVERPASS,NOGT]
COUNTRIES=["Faroe Islands","Guernsey","Isle of Man"]
OSM_COUNTRIES=["andorra","azores"]
IDSS=[(5164034601,5163680824),(920411192,920412407),(2197735231,2197735192)]
POLYGON=[54.4,10.4,54.5,10.4,54.5,10.5,54.4,10.5]
PACKAGES="requests,numpy,dogpile.cache,networkx,graph_tool,dbm.gnu,shapely,folium,flask,scipy,pyproj,setproctitle"

def install(package):
    from subprocess import call
    from sys import executable
    return call([executable, "-m", "pip", "-q", "install", package])
    
def install_packages(no_gt):
    from limic.util import start,end,status
    from importlib.util import find_spec
    packages = PACKAGES.split(",")
    for package in packages:
        if no_gt and package == "graph_tool":
            continue
        start("Checking for module",package)
        if find_spec(package):
            status("OK")
        else:
            status("MISSING")
            if package == "graph_tool":
                no_gt = True
            if package in ("graph_tool","dbm.gnu"):
                status("WARNING: "+package+"is needed for some optional functionality - if needed, it has to be installed manually")
                continue
            start("Trying to install",package,"using PIP")
            if install(package) == 0:
                end()
            else:
                status("FAILED")
                raise Exception("could not install",package,"using PIP - manual install?")
    return no_gt

def test(overpass_url,no_gt):
    no_gt = install_packages(no_gt)
    import limic.util as util
    import limic.init as init
    import limic.route as route
    import limic.convert as convert
    import limic.merge as merge
    import limic.prune as prune
    import limic.select as select
    import limic.length as length
    util.set_option('overwrite',True)
    countries = COUNTRIES
    osm_countries = OSM_COUNTRIES
    idss = IDSS
    init.init_stage0(overpass_url,countries,no_gt)
    init.init_stage1_cache(overpass_url,countries,no_gt)
    init.init_stage1_osm(osm_countries,no_gt)
    init.init_stage2(countries,no_gt)
    init.init_stage3(countries,no_gt)
    merge.merge_cache(list(map(lambda x:"cache."+x,countries)),"cache.merged")
    polygon = POLYGON
    for country, ids in zip(countries,idss):
        for suffix in ("nx","gt","npz"):
            if no_gt and suffix == "gt":
                continue
            graph_name = "graph."+country+"."+suffix
            for suffix2 in ("nx","gt","npz"):
                if suffix == suffix2 or (no_gt and suffix2 == "gt"):
                    continue
                graph_name2 = "graph."+country+"."+suffix2
                vars(convert)['convert_'+suffix+"_"+suffix2](file_name_in=graph_name,file_name_out=graph_name2)
            vars(route)['route_'+suffix](file_name=graph_name,source_id=ids[0],target_id=ids[1])
            vars(prune)['prune_'+suffix](file_name_in=graph_name,file_name_out="pruned."+country+"."+suffix,polygon=polygon,overpass_url=overpass_url)
            vars(select)['select_'+suffix](file_name_in=graph_name,file_name_out="selected."+country+"."+suffix,polygon=polygon,overpass_url=overpass_url)
            if not no_gt:
                vars(length)['length_'+suffix](file_name=graph_name)
        route.route_direct(file_name="cache."+country,source_id=ids[0],target_id=ids[1])



GRAPH = (("file_name_in",),{'type':str,'help':"use graph file GRAPH",'metavar':'GRAPH'})
HTML = (("file_name_out",),{'type':str,'help':"render to file HTML",'metavar':'HTML'})
MARKERS = (("--markers",),{'action':'store_true','dest':'markers','default':False,'help':"include markers (HUGE FILE)"})
LINES = (("--lines",),{'action':'store_true','dest':'lines','default':False,'help':"include lines (HUGE FILE)"})
HOST = (("--host",),{'type':str,'dest':'host','default':'localhost','help':"hostname (default: localhost)",'metavar':'HOST'})
PORT = (("--port",),{'type':int,'dest':'port','default':5000, 'help':"port number (default: 5000)",'metavar':'PORT'})
PREFIX = (("--prefix",),{'type':str,'dest':'prefix','default':'','help':"URI prefix (default: '')",'metavar':'PREFIX'})

CONFIG = [
    ("nx",{'help':"render HTML for NX graph",'args':[HOST,PORT,PREFIX,LINES,MARKERS,GRAPH,HTML]}),
    ("npz",{'help':"render HTML for NPZ graph",'args':[HOST,PORT,PREFIX,LINES,MARKERS,GRAPH,HTML]})
]

def render_nx(file_name_in,file_name_out,markers=False,lines=False,host="localhost",port=5000,prefix=""):
    from limic.util import start, end, load_pickled, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading NX graph",file_name_in)
    g = load_pickled(file_name_in)
    nodes = list(g.nodes())
    edges = list(g.edges.data('weight'))
    end()
    return render(g,nodes,edges,file_name_out,markers,lines,host,port,prefix)
    
def render_npz(file_name_in,file_name_out,markers=False,lines=False,host="localhost",port=5000,prefix=""):
    from limic.util import start, end, load_npz, check_overwrite
    from numpy import column_stack
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading NPZ graph",file_name_in)
    g = load_npz(file_name_in)
    nodes = list(map(tuple,column_stack((g['ids'],g['lat'],g['long']))))
    edges = []
    if lines:
        id2edges = g['id2edges']
        edges_weight = g['edges_weight']
        edges_neighbor = g['edges_neighbor']
        for index in range(len(nodes)):
            left = id2edges[index]
            right = id2edges[index+1]
            me = nodes[index]
            for i in range(left,right):
                edges.append((me,nodes[edges_neighbor[i]],edges_weight[i]))
    end()
    return render(g,nodes,edges,file_name_out,markers,lines,host,port,prefix)
    
def render(g,nodes,edges,file_name_out,markers=False,lines=False,host="localhost",port=5000,prefix=""):
    from limic.util import start, end, status, file_size
    from folium import Map, Marker, Icon, PolyLine
    from folium.plugins import BeautifyIcon
    from binascii import hexlify
    from pathlib import Path
    from math import log2
    from pkg_resources import resource_string
    start("Rendering graph")
    min_lat = min_long = float('inf')
    max_lat = max_long = -float('inf')
    for n in nodes:
        if n[1] < min_lat: min_lat = n[1]
        if n[2] < min_long: min_long = n[2]
        if n[1] > max_lat: max_lat = n[1]
        if n[2] > max_long: max_long = n[2]
    m = Map()
    m.fit_bounds([(min_lat,min_long),(max_lat,max_long)])
    if markers:
        for n in nodes:
            Marker(n[1:3],icon=BeautifyIcon(icon='none',                                       iconStyle="opacity: 0.1;",borderColor='#7f7f00',backgroundColor='#ffff00'),
               tooltip=("id: %d" % n[0])).add_to(m)
    if lines:
        for u,v,weight in edges:
            PolyLine([u[1:3],v[1:3]],color="#3f3f00",opacity=0.4,weight=6,tooltip=("weight: %.1f" % weight)).add_to(m)
    TEMPLATE = resource_string("limic","render.js").decode('utf8')
    from branca.element import MacroElement, Template
    class LatLngListener(MacroElement):
        _template = Template(TEMPLATE % {'host':host,'port':port,'prefix':prefix})
        def __init__(self):
            super(MacroElement, self).__init__()
            self._name = 'LatLngListener'
    LatLngListener().add_to(m)
    end()
    start("Saving result to HTML file",file_name_out)
    m.save(file_name_out)
    end('')
    file_size(file_name_out)
    return g,nodes



COUNTRIES = (("countries",),{'type':str,'nargs':'+','help':"countries to work on (PATTERN [w/o ANTIPATTERN])",'metavar':'COUNTRY'})
SUFFIXES = (("suffixes",),{'type':str,'nargs':'+','help':"suffixes to download",'metavar':'SUFFIX'})
URL = (("-d","--download-url"),{'type':str,'dest':'url','help':"define url for download directory to be URL",'metavar':'URL'})
SHOW = (("-l","--list"),{'action':'store_true','dest':'show','default':False,'help':"list available countries (default: False)"})
WORKERS = (("-w","--max-workers"),{'type':int,'dest':'max_workers','help':"set maximum number of parallel workers to WORKERS",'metavar':'WORKERS'})


CONFIG = [
    ("osm",{'help':"download map data",'args':[SHOW,URL,WORKERS,COUNTRIES]}),
    ("graph",{'help':"download graph files",'args':[
        ("nx",{'help':"download NX graph files",'args':[SHOW,URL,COUNTRIES]}),
        ("gt",{'help':"download GT graph files",'args':[SHOW,URL,COUNTRIES]}),
        ("npz",{'help':"download NPZ graph files",'args':[SHOW,URL,COUNTRIES]})
    ]}),
    ("cache",{'help':"download cache files from Overpass API",'args':[SHOW,URL,COUNTRIES]}),
    ("merged",{'help':"download merged graph file",'args':[URL,SUFFIXES]})
]
BASE_URL="http://caracal.imada.sdu.dk/d4e/"
COUNTRIES=['Albania','Andorra','Austria','Belarus','Belgium','Bosnia and Herzegovina','Bulgaria','Croatia','Cyprus','Czechia','Denmark','Estonia','Faroe Islands','Finland','France','Georgia','Germany','Gibraltar','Greece','Guernsey','Hungary','Iceland','Ireland','Isle of Man','Italy','Jersey','Kosovo','Latvia','Liechtenstein','Lithuania','Luxembourg','Macedonia','Malta','Moldova','Monaco','Montenegro','Netherlands','Norway','Poland','Portugal','RU','Romania','San Marino','Serbia','Slovakia','Slovenia','Spain','Sweden','Switzerland','Turkey','Ukraine','United Kingdom']
OSM_COUNTRIES=['africa','africa_algeria','africa_angola','africa_benin','africa_botswana','africa_burkina-faso','africa_burundi','africa_cameroon','africa_canary-islands','africa_cape-verde','africa_central-african-republic','africa_chad','africa_comores','africa_congo-brazzaville','africa_congo-democratic-republic','africa_djibouti','africa_egypt','africa_equatorial-guinea','africa_eritrea','africa_ethiopia','africa_gabon','africa_ghana','africa_guinea','africa_guinea-bissau','africa_ivory-coast','africa_kenya','africa_lesotho','africa_liberia','africa_libya','africa_madagascar','africa_malawi','africa_mali','africa_mauritania','africa_mauritius','africa_morocco','africa_mozambique','africa_namibia','africa_niger','africa_nigeria','africa_rwanda','africa_saint-helena-ascension-and-tristan-da-cunha','africa_sao-tome-and-principe','africa_senegal-and-gambia','africa_seychelles','africa_sierra-leone','africa_somalia','africa_south-africa','africa_south-africa-and-lesotho','africa_south-sudan','africa_sudan','africa_swaziland','africa_tanzania','africa_togo','africa_tunisia','africa_uganda','africa_zambia','africa_zimbabwe','antarctica','asia','asia_afghanistan','asia_armenia','asia_azerbaijan','asia_bangladesh','asia_bhutan','asia_cambodia','asia_china','asia_gcc-states','asia_india','asia_indonesia','asia_iran','asia_iraq','asia_israel-and-palestine','asia_japan','asia_japan_chubu','asia_japan_chugoku','asia_japan_hokkaido','asia_japan_kansai','asia_japan_kanto','asia_japan_kyushu','asia_japan_shikoku','asia_japan_tohoku','asia_jordan','asia_kazakhstan','asia_kyrgyzstan','asia_laos','asia_lebanon','asia_malaysia-singapore-brunei','asia_maldives','asia_mongolia','asia_myanmar','asia_nepal','asia_north-korea','asia_pakistan','asia_philippines','asia_south-korea','asia_sri-lanka','asia_syria','asia_taiwan','asia_tajikistan','asia_thailand','asia_turkmenistan','asia_uzbekistan','asia_vietnam','asia_yemen','australia-oceania','australia-oceania_australia','australia-oceania_cook-islands','australia-oceania_fiji','australia-oceania_marshall-islands','australia-oceania_micronesia','australia-oceania_new-caledonia','australia-oceania_new-zealand','australia-oceania_papua-new-guinea','australia-oceania_samoa','central-america','central-america_bahamas','central-america_belize','central-america_cuba','central-america_guatemala','central-america_haiti-and-domrep','central-america_jamaica','central-america_nicaragua','europe','europe_albania','europe_alps','europe_andorra','europe_austria','europe_azores','europe_belarus','europe_belgium','europe_bosnia-herzegovina','europe_britain-and-ireland','europe_bulgaria','europe_croatia','europe_cyprus','europe_czech-republic','europe_dach','europe_denmark','europe_estonia','europe_faroe-islands','europe_finland','europe_france','europe_france_alsace','europe_france_aquitaine','europe_france_auvergne','europe_france_basse-normandie','europe_france_bourgogne','europe_france_bretagne','europe_france_centre','europe_france_champagne-ardenne','europe_france_corse','europe_france_franche-comte','europe_france_guadeloupe','europe_france_guyane','europe_france_haute-normandie','europe_france_ile-de-france','europe_france_languedoc-roussillon','europe_france_limousin','europe_france_lorraine','europe_france_martinique','europe_france_mayotte','europe_france_midi-pyrenees','europe_france_nord-pas-de-calais','europe_france_pays-de-la-loire','europe_france_picardie','europe_france_poitou-charentes','europe_france_provence-alpes-cote-d-azur','europe_france_reunion','europe_france_rhone-alpes','europe_georgia','europe_germany','europe_germany_baden-wuerttemberg','europe_germany_baden-wuerttemberg_freiburg-regbez','europe_germany_baden-wuerttemberg_karlsruhe-regbez','europe_germany_baden-wuerttemberg_stuttgart-regbez','europe_germany_baden-wuerttemberg_tuebingen-regbez','europe_germany_bayern','europe_germany_bayern_mittelfranken','europe_germany_bayern_niederbayern','europe_germany_bayern_oberbayern','europe_germany_bayern_oberfranken','europe_germany_bayern_oberpfalz','europe_germany_bayern_schwaben','europe_germany_bayern_unterfranken','europe_germany_berlin','europe_germany_brandenburg','europe_germany_bremen','europe_germany_hamburg','europe_germany_hessen','europe_germany_mecklenburg-vorpommern','europe_germany_niedersachsen','europe_germany_nordrhein-westfalen','europe_germany_nordrhein-westfalen_arnsberg-regbez','europe_germany_nordrhein-westfalen_detmold-regbez','europe_germany_nordrhein-westfalen_duesseldorf-regbez','europe_germany_nordrhein-westfalen_koeln-regbez','europe_germany_nordrhein-westfalen_muenster-regbez','europe_germany_rheinland-pfalz','europe_germany_saarland','europe_germany_sachsen','europe_germany_sachsen-anhalt','europe_germany_schleswig-holstein','europe_germany_thueringen','europe_great-britain','europe_great-britain_england','europe_great-britain_england_bedfordshire','europe_great-britain_england_berkshire','europe_great-britain_england_bristol','europe_great-britain_england_buckinghamshire','europe_great-britain_england_cambridgeshire','europe_great-britain_england_cheshire','europe_great-britain_england_cornwall','europe_great-britain_england_cumbria','europe_great-britain_england_derbyshire','europe_great-britain_england_devon','europe_great-britain_england_dorset','europe_great-britain_england_durham','europe_great-britain_england_east-sussex','europe_great-britain_england_east-yorkshire-with-hull','europe_great-britain_england_essex','europe_great-britain_england_gloucestershire','europe_great-britain_england_greater-london','europe_great-britain_england_greater-manchester','europe_great-britain_england_hampshire','europe_great-britain_england_herefordshire','europe_great-britain_england_hertfordshire','europe_great-britain_england_isle-of-wight','europe_great-britain_england_kent','europe_great-britain_england_lancashire','europe_great-britain_england_leicestershire','europe_great-britain_england_lincolnshire','europe_great-britain_england_merseyside','europe_great-britain_england_norfolk','europe_great-britain_england_north-yorkshire','europe_great-britain_england_northamptonshire','europe_great-britain_england_northumberland','europe_great-britain_england_nottinghamshire','europe_great-britain_england_oxfordshire','europe_great-britain_england_rutland','europe_great-britain_england_shropshire','europe_great-britain_england_somerset','europe_great-britain_england_south-yorkshire','europe_great-britain_england_staffordshire','europe_great-britain_england_suffolk','europe_great-britain_england_surrey','europe_great-britain_england_tyne-and-wear','europe_great-britain_england_warwickshire','europe_great-britain_england_west-midlands','europe_great-britain_england_west-sussex','europe_great-britain_england_west-yorkshire','europe_great-britain_england_wiltshire','europe_great-britain_england_worcestershire','europe_great-britain_scotland','europe_great-britain_wales','europe_greece','europe_hungary','europe_iceland','europe_ireland-and-northern-ireland','europe_isle-of-man','europe_italy','europe_italy_centro','europe_italy_isole','europe_italy_nord-est','europe_italy_nord-ovest','europe_italy_sud','europe_kosovo','europe_latvia','europe_liechtenstein','europe_lithuania','europe_luxembourg','europe_macedonia','europe_malta','europe_moldova','europe_monaco','europe_montenegro','europe_netherlands','europe_netherlands_drenthe','europe_netherlands_flevoland','europe_netherlands_friesland','europe_netherlands_gelderland','europe_netherlands_groningen','europe_netherlands_limburg','europe_netherlands_noord-brabant','europe_netherlands_noord-holland','europe_netherlands_overijssel','europe_netherlands_utrecht','europe_netherlands_zeeland','europe_netherlands_zuid-holland','europe_norway','europe_poland','europe_poland_dolnoslaskie','europe_poland_kujawsko-pomorskie','europe_poland_lodzkie','europe_poland_lubelskie','europe_poland_lubuskie','europe_poland_malopolskie','europe_poland_mazowieckie','europe_poland_opolskie','europe_poland_podkarpackie','europe_poland_podlaskie','europe_poland_pomorskie','europe_poland_slaskie','europe_poland_swietokrzyskie','europe_poland_warminsko-mazurskie','europe_poland_wielkopolskie','europe_poland_zachodniopomorskie','europe_portugal','europe_romania','europe_serbia','europe_slovakia','europe_slovenia','europe_spain','europe_sweden','europe_switzerland','europe_turkey','europe_ukraine','north-america','north-america_canada','north-america_canada_alberta','north-america_canada_british-columbia','north-america_canada_manitoba','north-america_canada_new-brunswick','north-america_canada_newfoundland-and-labrador','north-america_canada_northwest-territories','north-america_canada_nova-scotia','north-america_canada_nunavut','north-america_canada_ontario','north-america_canada_prince-edward-island','north-america_canada_quebec','north-america_canada_saskatchewan','north-america_canada_yukon','north-america_greenland','north-america_mexico','north-america_us','north-america_us-midwest','north-america_us-northeast','north-america_us-pacific','north-america_us-south','north-america_us-west','north-america_us_alabama','north-america_us_alaska','north-america_us_arizona','north-america_us_arkansas','north-america_us_california','north-america_us_california_norcal','north-america_us_california_socal','north-america_us_colorado','north-america_us_connecticut','north-america_us_delaware','north-america_us_district-of-columbia','north-america_us_florida','north-america_us_georgia','north-america_us_hawaii','north-america_us_idaho','north-america_us_illinois','north-america_us_indiana','north-america_us_iowa','north-america_us_kansas','north-america_us_kentucky','north-america_us_louisiana','north-america_us_maine','north-america_us_maryland','north-america_us_massachusetts','north-america_us_michigan','north-america_us_minnesota','north-america_us_mississippi','north-america_us_missouri','north-america_us_montana','north-america_us_nebraska','north-america_us_nevada','north-america_us_new-hampshire','north-america_us_new-jersey','north-america_us_new-mexico','north-america_us_new-york','north-america_us_north-carolina','north-america_us_north-dakota','north-america_us_ohio','north-america_us_oklahoma','north-america_us_oregon','north-america_us_pennsylvania','north-america_us_puerto-rico','north-america_us_rhode-island','north-america_us_south-carolina','north-america_us_south-dakota','north-america_us_tennessee','north-america_us_texas','north-america_us_utah','north-america_us_vermont','north-america_us_virginia','north-america_us_washington','north-america_us_west-virginia','north-america_us_wisconsin','north-america_us_wyoming','russia','russia_central-fed-district','russia_crimean-fed-district','russia_far-eastern-fed-district','russia_kaliningrad','russia_north-caucasus-fed-district','russia_northwestern-fed-district','russia_siberian-fed-district','russia_south-fed-district','russia_ural-fed-district','russia_volga-fed-district','south-america','south-america_argentina','south-america_bolivia','south-america_brazil','south-america_brazil_centro-oeste','south-america_brazil_nordeste','south-america_brazil_norte','south-america_brazil_sudeste','south-america_brazil_sul','south-america_chile','south-america_colombia','south-america_ecuador','south-america_paraguay','south-america_peru','south-america_suriname','south-america_uruguay','south-america_venezuela']
#OSM_COUNTRIES=['albania','andorra','austria','azores','belarus','belgium','bosnia-herzegovina','bulgaria','croatia','cyprus','czech-republic','denmark','estonia','faroe-islands','finland','france','georgia','germany','great-britain','greece','hungary','iceland','ireland-and-northern-ireland','isle-of-man','italy','kosovo','latvia','liechtenstein','lithuania','luxembourg','macedonia','malta','moldova','monaco','montenegro','netherlands','norway','poland','portugal','romania','russia','serbia','slovakia','slovenia','spain','sweden','switzerland','turkey','ukraine']
OSM_URL="https://download.geofabrik.de/"

def download_file(url,file_name,retries=10):
    from requests import get
    from shutil import copyfileobj
    from sys import exit
    from limic.util import status, md5file, options
    from os.path import exists
    download = True
    if options.md5sum:
        file_md5 = md5file(file_name) if exists(file_name) else None
        md5_retries = retries
        while True:
            timed_out = False
            try:
                r = get(url+".md5",timeout=5)
            except:
                timed_out = True
            if not timed_out and r.status_code == 200:
                break
            md5_retries -= 1
            if not md5_retries:
                status("ERROR(file.md5): HTTP status 200 expected, got "+str(r.status_code)+" for "+url+".md5")
                exit(-1)
            status("WARNING(download): RETRYING "+url+".md5")
        url_md5 = r.content.split()[0].decode('utf8')
        if file_md5 and url_md5 == file_md5:
            if options.verbosity >= 2:
                status('SKIPPING',end='   ')
            return file_name
    while True:
        r = get(url, stream=True)
        if r.status_code == 200:
            break
        retries -= 1
        if not retries:
            status("ERROR(file): HTTP status 200 expected, got "+str(r.status_code)+" for "+url)
            exit(-1)
        status("WARNING(download): RETRYING "+url+".md5")
    f = open(file_name, 'wb')
    copyfileobj(r.raw, f)
    f.close()
    if options.md5sum:
        file_md5 = md5file(file_name)
        if url_md5 != file_md5:
            status("ERROR(md5): "+file_md5+" vs "+url_md5)
            exit(-1)
        f = open(file_name+".md5","wt")
        f.write(file_md5+" "+file_name)
        f.close()
    return file_name

def download_osm(countries,url=None,show=False,max_workers=None):
    from limic.util import start,end,file_size,status
    from concurrent.futures import ProcessPoolExecutor, wait
    countries, url = common(countries,url,show,osm=True)
    if max_workers:
        executor = ProcessPoolExecutor(max_workers=max_workers)
        fs = []
        start("Downloading OSM map data for"," ".join(countries))
    for country in countries:
        file_url = OSM_URL+country.replace("_","/")+"-latest.osm.bz2"
        file_name = country+"-latest.osm.bz2"
        if max_workers:
            fs.append(executor.submit(download_file,file_url,file_name))
            continue
        start("Downloading OSM map data for",country)
        file_name = download_file(file_url,file_name)
        end('')
        file_size(file_name)
    if max_workers:
        running = len(fs)
        total = running
        while running:
            print("Waiting for",running,"out of",total,"processes ...")
            wait(fs,timeout=10)
            running = sum(0 if f.done() else 1 for f in fs)
        for f in fs:
            file_name = f.result()
            status(file_name,end=':')
            file_size(file_name,end='   ')
        end()

def download_cache(countries,url=None,show=False):
    from limic.util import start,end,file_size
    countries, url = common(countries,url,show)
    for country in countries:
        start("Downloading cache for",country)
        file_url = url+"cache."+country.replace(" ","%20")
        file_name = download_file(file_url,"cache."+country)
        end('')
        file_size(file_name)

def download_graph(suffix,countries,url=None,show=False,join=False):
    from limic.util import start,end,file_size
    countries, url = common(countries,url,show=show,join=join)
    for country in countries:
        start("Downloading",suffix.upper(),"graph for",country)
        file_url = url+"graph."+country.replace(" ","%20")+"."+suffix
        file_name = download_file(file_url,"graph."+country+"."+suffix)
        end('')
        file_size(file_name)

def download_graph_nx(countries,url=None,show=False):
    download_graph("nx",countries,url=url,show=show)

def download_graph_gt(countries,url=None,show=False):
    download_graph("gt",countries,url=url,show=show)

def download_graph_npz(countries,url=None,show=False):
    download_graph("npz",countries,url=url,show=show)
        
def download_merged(suffixes=("nx","gt","npz"),url=None):
    from limic.util import start,end,file_size
    for suffix in suffixes:
        start("Downloading merged",suffix.upper(),"graph for Europe")
        file_url = url+"merged.Europe."+suffix
        file_name = download_file(file_url,"merged.Europe."+suffix)
        end('')
        file_size(file_name)

def common(countries,url=None,show=False,osm=False,join=False):
    from fnmatch import translate
    from re import compile
    selected = set()
    anti = False
    ALL_COUNTRIES = OSM_COUNTRIES+COUNTRIES if join else (OSM_COUNTRIES if osm else COUNTRIES)
    for country in countries:
        if country == 'w/o':
            anti = True
        country = compile(translate(country).replace("%%",".*").replace("%","[^_]*"))
        for available in ALL_COUNTRIES:
            if country.match(available):
                if anti:
                    selected.discard(available)
                else:
                    selected.add(available)
    countries = list(selected)
    countries.sort()
    if show or not countries:
        from limic.util import options
        options.parser.error("Available countries:\n"+" ".join(ALL_COUNTRIES))
    if not url:
        url = OSM_URL if osm else BASE_URL
    return countries, url



CACHE = (("file_name_in",),{'type':str,'help':"read map data from cache file CACHE",'metavar':'CACHE'})
LIM = (("file_name_in",),{'type':str,'help':"read map data from LiMiC file FILE",'metavar':'LIM'})
OSM = (("file_name_in",),{'type':str,'help':"read map data from OpenStreetMap file OSM",'metavar':'OSM'})
OUT = (("file_name_out",),{'type':str,'help':"save extracted graph to file GRAPH",'metavar':'GRAPH'})
EPS = (("-e","--epsilon"),{'type':float,'dest':'eps','default':0.01,'help':"set minimum percentage of  linear infrastructure segment to consider non-logical intersections to EPS",'metavar':'EPS'})
AROUND = (("-r","--around"),{'type':int,'dest':'around','default':1000,'help':"set maximum free flying distance in m around linear infrastructure nodes to AROUND",'metavar':'AROUND'})
SAFEDIST = (("-s","--safe-dist"),{'type':int,'dest':'safe_dist','default':100,'help':"set minimum safe distance from blacklisted areas and structures to SAFE",'metavar':'SAFE'})
PENALIZE = (("-p","--penalize"),{'type':int,'default':20,'dest':'penalize','help':"penalize free flying disrances by factor PENALTY",'metavar':'PENALTY'})
WHITE = (("--white",),{'type':str,'dest':'white','default':'{"power":"line"}','help':"linear infrastructure to include (default: {'power':'line'})",'metavar':'WHITELIST'})
BLACK = (("--black",),{'type':str,'dest':'black','default':'{"power":"substation"}','help':"areas and structures to avoid (default: {'power':'substation'})",'metavar':'BLACKLIST'})
CONSERVEMEM = (("-c","--conserve-memory"),{'action':'store_true','dest':'conserve_mem','default':False,'help':"lower memory usage but higher runtime"})

CONFIG = [
    ("cache",{'args':[
        (("-u","--overpass-url"),{'type':str,'dest':'overpass_url','help':"define url for Overpass API to be URL",'metavar':'URL'}),
        (("-a","--area"),{'type':str,'dest':'area','help':"set area to extract to AREA",'metavar':'AREA'}),
        AROUND,EPS,SAFEDIST,PENALIZE,CACHE,OUT
    ]}),
    ("osm",{'args':[BLACK,WHITE,CONSERVEMEM,AROUND,EPS,SAFEDIST,PENALIZE,OSM,OUT]}),
    ("osm_pre",{'args':[BLACK,WHITE,CONSERVEMEM,OSM,OUT]}),
    ("osm_post",{'args':[AROUND,EPS,SAFEDIST,PENALIZE,LIM,OUT]})
    #("esy",{'args':[OSM,OUT]})
]

def prune_incomplete(g):
    to_prune = set()
    for k in g.nodes():
        gk = g[k]
        for l in g.neighbors(k):
            if k < l and gk[l]['type'] < 0:
                gl = g[l]
                for m in g.neighbors(l):
                    if g.has_edge(k,m) and gk[l]['weight'] > gk[m]['weight']+gl[m]['weight']:
                        to_prune.add((k,l))
    for u,v in to_prune:
        if g.has_edge(u,v):
            g.remove_edge(u,v)

def prune_complete(g):
    from networkx import astar_path_length
    to_prune = set()
    for k in g.nodes():
        for l in g.neighbors(k):
            if k < l:
                gkl = g[k][l]
                if gkl['type'] < 0 and gkl['weight'] > astar_path_length(g,k,l):
                    to_prune.add((k,l))
    for u,v in to_prune:
        if g.has_edge(u,v):
            g.remove_edge(u,v)

def build_edges(g,find_all_neighbours,around,eps,safe_dist,penalize):
    from limic.overpass import pylon
    from limic.util import distance
    #count = 0
    neighbours2intersection = {}
    minusid = [0]
    latlon2id = {}
    for tower in list(g.nodes()):
        #count += 1
        #if verbosity >= 2: print(count,"of",total)
        for neighbour in find_all_neighbours(tower,around,eps,safe_dist,minusid,latlon2id,penalize):
            if neighbour.tower in g or neighbour.tower.id < 0:
                #if verbosity >= 2: print("adding",neighbour)
                g.add_edge(tower,pylon(neighbour.tower.id,neighbour.tower.latlon),weight=neighbour.dist,type=neighbour.air)
                if neighbour.tower.id < 0:
                    for intersection in neighbours2intersection.setdefault(neighbour.tower.neighbours,[]):
                        if intersection.latlon != neighbour.tower.latlon:
                            #if verbosity >= 2: print("double intersection:",intersection.latlon,neighbour.tower.latlon)
                            g.add_edge(intersection,neighbour.tower,weight=distance(intersection.latlon,neighbour.tower.latlon),type=False)
                    neighbours2intersection[neighbour.tower.neighbours].append(neighbour.tower)


def extract_cache(file_name_in,file_name_out,overpass_url,area=None,around=1000,eps=0.01,safe_dist=100,penalize=20):
    from limic.overpass import distance, find_all_neighbours, is_safe, set_server, pylon, region, get_towers_by_area
    from limic.util import start, end, file_size, status, load_pickled, save_pickled, replace, check_overwrite
    from networkx import Graph
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading",file_name_in)
    region.backend._cache = load_pickled(file_name_in)
    len_cache = len(region.backend._cache)
    end('')
    file_size(file_name_in)
    if not area:
        area = file_name_in.split(".")[1]
    start("Querying overpass for",area)
    set_server(overpass_url)
    towers = get_towers_by_area(area)
    end()
    start("Building safe nodes")
    g=Graph()
    for tower in towers:
        if is_safe(tower,safe_dist):
            g.add_node(tower)
#        else:
#        if verbosity >= 2: print("NOT safe!")
    end('')
    total = len(g.nodes())
    status(total)
    start("Building edges")
    build_edges(g,find_all_neighbours,around,eps,safe_dist,penalize)
    end('')
    status(len(g.edges()))
    if len_cache != len(region.backend._cache):
        file_name_tmp = file_name_in+".tmp"
        start("Saving to",file_name_in,"via",file_name_tmp)
        save_pickled(file_name_tmp,region.backend._cache)
        replace(file_name_tmp,file_name_in)
        end('')
        file_size(file_name_in)
    from limic.util import start, end, status, file_size, save_pickled
    from networkx import relabel_nodes
    start("Prune redundant edges (incomplete)")
    prune_incomplete(g)
    end('')
    status(len(g.edges()))
    start("Prune redundant edges (complete)")
    prune_complete(g)
    end('')
    status(len(g.edges()))
    start("Cleaning up graph")
    relabel = dict(map(lambda tower:(tower,(tower.id,tower.latlon[0],tower.latlon[1])),g.nodes()))
    relabel_nodes(g,relabel,copy=False)
    for u,v,d in g.edges(data=True):
        d['type'] = -1 if d['type'] else 0
    end()
    start("Saving graph to",file_name_out)
    save_pickled(file_name_out,g)
    end('')
    file_size(file_name_out)

def extract_osm(file_name_in,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20,white="{'power':'line'}",black="{'power':'substation'}",conserve_mem=False):
    from setproctitle import setproctitle
    setproctitle(file_name_in)
    from limic.util import check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    white,black = list(eval(white).items()), list(eval(black).items())
    lim = osm_pre(file_name_in,white,black,conserve_mem)
    osm_post(lim,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20)

def extract_osm_pre(file_name_in,file_name_out,white="{'power':'line'}",black="{'power':'substation'}",conserve_mem=False):
    from limic.util import start, end, save_pickled, file_size, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    white,black = list(eval(white).items()), list(eval(black).items())
    lim = osm_pre(file_name_in,white,black,conserve_mem)
    start("Saving data to",file_name_out)
    save_pickled(file_name_out,lim)
    end('')
    file_size(file_name_out)

def extract_osm_post(file_name_in,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20):
    from limic.util import start, end, file_size, load_pickled, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading filtered OSM data from",file_name_in)
    lim = load_pickled(file_name_in)
    end('')
    file_size(file_name_in)
    osm_post(lim,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20)

def osm_post(lim,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20):
    from limic.util import start, end, status, file_size, load_pickled, distance, save_pickled
    from scipy.spatial import cKDTree as KDTree
    from networkx import Graph, astar_path_length
    from pyproj import CRS, Transformer
    from itertools import chain
    from limic.overpass import intersect, pylon
    lines, substations, towers, id2tower, id2node, id2lines, id2types = lim
    start("Building KD-tree from white nodes")
    from limic.util import kdtree
    towers_tree = kdtree(towers,get_latlon=lambda x:x.latlon)
    end('')
    status(len(towers))
    start("Deleting black nodes")
    to_delete = set()
    from limic.util import nodes_in_geometry
    for substation in substations:
        to_delete.update(nodes_in_geometry(towers_tree,list(map(lambda x:id2node[x],substation))))
    towers = [tower for tower in towers if tower not in to_delete]
    end('')
    status(len(towers))
    start("Building initial graph")
    g = Graph()
    g.add_nodes_from(towers)
    for line in lines:
        line_nodes = list(map(lambda x:id2tower[x],line))
        for from_node, to_node in zip(line_nodes,line_nodes[1:]):
            if from_node in to_delete or to_node in to_delete:
                continue
            w = distance(from_node.latlon,to_node.latlon)
            g.add_edge(from_node,to_node,weight=w,type=id2types[from_node.id])
    end('')
    status(len(g.nodes()),end='/')
    status(len(g.edges()))
    start("Finding neighbours within "+str(around)+"m")
    towers_tree = kdtree(towers,get_latlon=lambda x:x.latlon)
    end('')
    neighbour_indices, neighbours = towers_tree.get_neighbours(around=1000)
    end()
    start("Computing non-logical intersections")
    tower2index = {}
    for i,t in zip(range(len(towers)),towers):
        tower2index[t] = i
    for k,v in id2lines.items():
        id2lines[k] = tuple(map(tuple,v))
    end('')
    segments = set()
    for u,v in g.edges():
        this = (u,v) if u < v else (v,u)
        ui, vi = tower2index[u], tower2index[v]
        lines = set()
        lines.update(id2lines[u.id])
        lines.update(id2lines[v.id])
        for neighbour in chain(neighbours[ui],neighbours[vi]):
            if neighbour == u or neighbour == v:
                continue
            if not lines.intersection(id2lines[neighbour.id]):
                for nn in g.neighbors(neighbour):
                    other = (neighbour,nn) if neighbour < nn else (nn,neighbour)
                    segments.add(tuple(sorted((this,other))))
    end('')
    status(len(segments),end='   ')
    neighbours2intersection = {}
    minusid = 0
    latlon2id = {}
    segments2intersections = {}
    for (t1,t2),(t3,t4) in segments:
        res = intersect(t1.latlon,t2.latlon,t3.latlon,t4.latlon,eps=eps,no_tu=False)
        if res:
            intersection,(t,u) = res
            if not intersection in latlon2id:
                minusid -= 1
                latlon2id[intersection] = minusid
            segments2intersections.setdefault((t1,t2),[]).append((t,latlon2id[intersection],intersection))
            segments2intersections.setdefault((t3,t4),[]).append((u,latlon2id[intersection],intersection))
    end('')
    status(-minusid,end='   ')
    for (u,v), intersections in segments2intersections.items():
        intersections.sort()
        g.remove_edge(u,v)
        type = id2types[u.id]
        assert(type == id2types[v.id])
        seq = [u]
        for _,id,latlon in intersections:
            seq.append(pylon(id,latlon))
        seq.append(v)
        for from_node, to_node in zip(seq,seq[1:]):
            w = distance(from_node.latlon,to_node.latlon)
            g.add_edge(from_node,to_node,weight=w,type=type)
    end()
    start("Adding routing through air")
    airs = set()
    for ns in neighbours:
        n = ns[0]
        for m in ns[1:]:
            if not g.has_edge(n,m):
                airs.add((n,m))
    end('')
    for n,m in airs:
        w = penalize*distance(n.latlon,m.latlon)
        g.add_edge(n,m,weight=w,type=-1)
    end('')
    status(len(g.nodes()),end='/')
    status(len(g.edges()))
    from networkx import relabel_nodes
    start("Prune redundant edges (incomplete)")
    prune_incomplete(g)
    end('')
    status(len(g.edges()))
    start("Prune redundant edges (complete)")
    prune_complete(g)
    end('')
    status(len(g.edges()))
    start("Cleaning up graph")
    relabel = dict(map(lambda tower:(tower,(tower.id,tower.latlon[0],tower.latlon[1])),g.nodes()))
    relabel_nodes(g,relabel,copy=False)
    end()
    start("Saving graph to",file_name_out)
    save_pickled(file_name_out,g)
    end('')
    file_size(file_name_out)

def osm_pre(file_name_in,white=[("power","line")],black=[("power","substation")],conserve_mem=False):
    from limic.util import start, end, status, file_size, split, file_copy
    from bz2 import open as bopen
    from limic.overpass import pylon
    white = list(map(lambda x:(x[0],x[1][0]+'" v="'+x[1][1]+'"'),zip(range(len(white)),white)))
    black = list(map(lambda x:(-1,x[0]+'" v="'+x[1]+'"'),black))
    white_black = white+black
    if not conserve_mem:
        start("Reading map data from",file_name_in)
        from mmap import mmap
        area = mmap(-1,2**40)
        if file_name_in[-4:] == ".bz2":
            try:
                from subprocess import Popen, PIPE
                f = Popen(['lbzcat',file_name_in],stdout=PIPE).stdout
            except Exception as e:
                print(e)
                status("WARNING: lbzcat failed - falling back on Python bz2 library")
                f = bopen(file_name_in,"rb")
        else:
            f = open(file_name_in,"rb")
        file_copy(f,area)
        f.close()
        area_length = area.tell()
        end()
    start("Extracting ways for white and black")
    if conserve_mem:
        f = bopen(file_name_in,"rb") if file_name_in[-4:] == ".bz2" else open(file_name_in,"rb")
        elems = split(f,'\n  <')
    else:
        area.seek(0)
        elems = split(area,'\n  <',maxlength=area_length)
    lines = []
    substations = []
    id2types = {}
    for elem in elems:
        if elem.startswith('way id="'):
            found = False
            for i,t in white_black:
                if t in elem:
                    found = True
            if not found:
                continue
            tags = elem.split('\n    <tag k="')
            for tag in tags[1:]:
                for i,t in white_black:
                    if tag.startswith(t):
                        nds = list(map(lambda x:int(x.split('"')[0]),tags[0].split('\n    <nd ref="')[1:]))
                        if i >= 0:
                            lines.append(nds)
                            for nd in nds:
                                id2types[nd] = i
                        else:
                            substations.append(nds)
    end('')
    status(len(lines),end='   ')
    status(len(substations))
    id2lines, id2substations = {}, {}
    for line in lines:
        for node in line:
            id2lines.setdefault(node,[]).append(line)
    for substation in substations:
        for node in substation:
            id2substations.setdefault(node,[]).append(substation)
    start("Extracting nodes for white and black")
    if conserve_mem:
        f = bopen(file_name_in,"rb") if file_name_in[-4:] == ".bz2" else open(file_name_in,"rb")
        elems = split(f,'\n  <')
    else:
        area.seek(0)
        elems = split(area,'\n  <',maxlength=area_length)
    towers = []
    id2tower = {}
    id2node = {}
    ids = set()
    ids.update(map(lambda x:str(x),id2lines.keys()))
    ids.update(map(lambda x:str(x),id2substations.keys()))
    for elem in elems:
        if elem.startswith('node id="'):
            if not elem[9:elem.find('"',10)] in ids:
                continue
            parts = elem.split("\n")[0].split('"')
            assert(parts[-5]==' lat=')
            assert(parts[-3]==' lon=')
            i = int(parts[1])
            if i in id2lines:
                elem = pylon(i,(float(parts[-4]),float(parts[-2])))
                towers.append(elem)
                id2tower[i] = elem
            if i in id2substations:
                elem = (float(parts[-4]),float(parts[-2]))
                id2node[i] = elem
    end('')
    status(len(towers))
    start("Checking all nodes were extracted")
    assert(set(id2lines.keys())==set(map(lambda x:x.id,towers)))
    end()
    return lines, substations, towers, id2tower, id2node, id2lines, id2types



from requests import get
from limic.util import distance, options
from dogpile.cache import make_region

server = "http://caracal.imada.sdu.dk/overpass"
def set_server(base_url):
    if base_url is not None:
        global server
        server = base_url

def key_generator(namespace, fn):
    namespace = "__main__:%s" % fn.__name__
    def generate_key(*args, **kw):
        if kw: raise ValueError("dogpile.cache's default key creation function does not accept keyword arguments.")
        return namespace + "|" + " ".join(map(str, args))
    return generate_key
if options.disk_cache:
    region = make_region(function_key_generator=key_generator).configure("dogpile.cache.dbm",arguments={"filename":options.disk_cache})
else:
    region = make_region(function_key_generator=key_generator).configure("dogpile.cache.memory")

class pylon:
    def __init__(self,id,latlon,neighbours=()):
        self.id = id
        self.latlon = latlon
        self.neighbours = neighbours
    def __repr__(self):
        return "pylon(id="+repr(self.id)+",latlon="+repr(self.latlon)+", neighbours="+repr(self.neighbours)+")"
    def __lt__(self,other):
        return self.latlon < other.latlon
    def __eq__(self,other):
        return isinstance(other,pylon) and self.latlon == other.latlon
    def __hash__(self):
        return hash(self.latlon)

class dist:
    def __init__(self,dist,air,tower):
        self.dist = dist
        self.air = air
        self.tower = tower
    def __lt__(self,other):
        return self.dist < other.dist
    def __repr__(self):
        return "dist(dist="+repr(self.dist)+",air="+repr(self.air)+",tower="+repr(self.tower)+")"

def overpass(query,timeout=3600,retries=10):
    from limic.util import options, status
    url = server+"/api/interpreter?data=%5Bout:json%5D%5Btimeout:3600%5D;"+query
    #if verbosity >= 2: print(url)
    while retries:
        try:
            r = get(url,timeout=timeout)
            return r.json()
        except Exception as exc:
            retries -= 1
            #if verbosity >= 4: print(exc)
            status("WARNING(overpass): RETRYING "+url)
    options.failed = True

def tower_by_id(tower_id):
    jsonfile = overpass("node(id:"+str(tower_id)+");out;")
    for child in jsonfile['elements']:
        tower = pylon(child['id'],(float(child['lat']),float(child['lon'])))
        return tower

@region.cache_on_arguments()
def find_towers(location, around):
    #from util import verbosity
    towers = []
    #if verbosity >= 2: print("finding towers close to",location)
    jsonfile = overpass("(node(around:"+str(around)+","+str(location[0])+","+str(location[1])+")%5B%22power%22=%22tower%22%5D;);out;")
    for child in jsonfile['elements']:
        tower = pylon(child['id'],(float(child['lat']),float(child['lon'])))
        towers.append(tower)
    #if verbosity >= 2: print("found",len(towers),"towers:",towers)
    return towers

@region.cache_on_arguments()
def find_segments(location, around):
    #from util import verbosity
    segments = []
    #if verbosity >= 2: print("finding lines close to",location)
    jsonfile = overpass("(way(around:"+str(around)+","+str(location[0])+","+str(location[1])+")%5B%22power%22=%22line%22%5D;node(w););out;")
    id2pos = {}
    for child in jsonfile['elements']:
        if child['type'] == "node":
            id2pos[child['id']] = (child['lat'],child['lon'])
        if child['type'] == "way":
            nodes = child['nodes']
            for i in range(1,len(nodes)):
                segments.append((nodes[i-1],nodes[i]))
    #if verbosity >= 2: print("found",len(segments),"segments:",segments)
    segments = list(map(lambda ids:(pylon(ids[0],id2pos[ids[0]]),pylon(ids[1],id2pos[ids[1]])),segments))
    return segments

def find_neighbours(tower):
    #from util import verbosity
    if tower.id < 0:
        towers = set()
        towers.update(tower.neighbours)
        #if verbosity >= 2: print("found",len(towers),"pre-defined direct neighbours:",towers)
        return towers
    return find_neighbours_id(tower)

@region.cache_on_arguments()
def find_neighbours_id(tower):
    #from util import verbosity
    #if verbosity >= 2: print("finding neighbours for",tower)
    towers = set()
    jsonfile = overpass("(node(id:"+str(tower.id)+");way(bn)%5B%22power%22=%22line%22%5D;node(w););out;")

    id2pos = {}
    for child in jsonfile['elements']:
        if child['type'] == "node" :
            id2pos[child['id']] = (child['lat'],child['lon'])
        if child['type'] == "way":
            nodes = child['nodes']
            index = nodes.index(tower.id)
            if index > 0:
                towers.add(nodes[index-1])
            if index+1 < len(nodes):
                towers.add(nodes[index+1])
    towers = list(map(lambda id:pylon(id,id2pos[id]),towers))
    #if verbosity >= 2: print("found",len(towers),"direct neighbours:",towers)
    return towers

@region.cache_on_arguments()
def around_station(location, around):
    #from util import verbosity
    #if verbosity >= 2: print("avoiding power stations less than",around,"metres from",location)
    jsonfile = overpass("(way(around:"+str(around)+","+str(location[0])+","+str(location[1])+")%5B%22power%22=%22substation%22%5D;);out;")
    for child in jsonfile['elements']:
        if child['type'] == "way":
            return True
    return False

@region.cache_on_arguments()
def find_substations(location,around):
    #from util import verbosity
    #if verbosity >= 2: print("retrieving power station geometries",around,"around",location)
    jsonfile = overpass("way(around:"+str(around)+","+str(location[0])+","+str(location[1])+")%5B%22power%22=%22substation%22%5D;out geom;")
    substations = []
    for child in jsonfile['elements']:
        if child['type'] == "way":
            geometry = list(map(lambda loc:(loc['lat'],loc['lon']),child['geometry']))
            substations.append(geometry)
    #if verbosity >= 2: print("substations:",substations)
    return substations

@region.cache_on_arguments()
def is_in(tower,geometry):
    #from util import verbosity
    #if verbosity >= 2: print("checking whether",tower,"is enclosed by",geometry)
    if tower.id < 0:
        #if verbosity >= 2: print("tower safe (is intersection):",tower)
        return False
    if len(geometry) < 3:
        #if verbosity >= 2: print("tower safe (unusable geometry):",tower)
        return False
    if len(geometry) > 100:
        #if verbosity >= 2: print("tower NOT safe (too large geometry")
        return True
    return tower.id in nodes_in_geometry(geometry)

def is_safe(tower, around):
    if not around_station(tower.latlon,around):
        return True
    for substation in find_substations(tower.latlon, around):
        if is_in(tower,substation):
            return False
    return True

def dist_towers(towers, location):
    d_towers = []
    for tower in towers:
        d = distance(location,tower.latlon)
        d_towers.append(dist(d,False,tower))
    return d_towers

def intersect(p1,p2,p3,p4,eps,no_tu=True):
    d = (p1[0]-p2[0])*(p3[1]-p4[1])-(p1[1]-p2[1])*(p3[0]-p4[0])
    if d == 0:
        return None
    t = ((p1[0]-p3[0])*(p3[1]-p4[1])-(p1[1]-p3[1])*(p3[0]-p4[0]))/d
    if t < eps or t > 1-eps:
        return None
    u = -((p1[0]-p2[0])*(p1[1]-p3[1])-(p1[1]-p2[1])*(p1[0]-p3[0]))/d
    if u < eps or u > 1-eps:
        return None
    x = ((p1[0]*p2[1]-p1[1]*p2[0])*(p3[0]-p4[0])-(p1[0]-p2[0])*(p3[0]*p4[1]-p3[1]*p4[0]))/d
    y = ((p1[0]*p2[1]-p1[1]*p2[0])*(p3[1]-p4[1])-(p1[1]-p2[1])*(p3[0]*p4[1]-p3[1]*p4[0]))/d
    return (x,y) if no_tu else (x,y),(t,u)
        
def find_all_neighbours(tower, around, eps, safe_dist, minusid, latlon2id, penalty):
    #from util import verbosity
    direct_neighbours = find_neighbours(tower)
    neighbours = dist_towers(direct_neighbours,tower.latlon)
    direct_segments = list(map(lambda neighbour:(neighbour.tower,tower),neighbours))
    indirect_neighbours = []
    for neighbour in find_towers(tower.latlon, around):
        if not neighbour in direct_neighbours and neighbour.id != tower.id:
             neighbours.append(dist(penalty*distance(neighbour.latlon,tower.latlon),True,neighbour))
        #else:
             #if verbosity >= 3: print("skipping direct neighbour/self",neighbour)
    indirect_segments = find_segments(tower.latlon, around)
    for indirect_segment in indirect_segments:
        for direct_segment in direct_segments:
            intersection = intersect(direct_segment[0].latlon,direct_segment[1].latlon,indirect_segment[0].latlon,indirect_segment[1].latlon,eps)
            if intersection:
                #if verbosity >= 3:
                #    print("found intersection",direct_segment,indirect_segment,intersection)
                #    print("node(id:"+(",".join(map(lambda t:str(t.id),[direct_segment[0],direct_segment[1],indirect_segment[0],indirect_segment[1]])))+");out;")
                if not intersection in latlon2id:
                    minusid[0] -= 1
                    latlon2id[intersection] = minusid[0]
                neighbours.append(dist(distance(intersection,tower.latlon),False,pylon(latlon2id[intersection], intersection,neighbours=indirect_segment)))
    #if verbosity >= 2: print("found a total of ",len(neighbours),"neighbours:",neighbours)
    safe_neighbours = []
    for neighbour in neighbours:
        if is_safe(neighbour.tower, safe_dist):
            safe_neighbours.append(neighbour)
        #else:
            #if verbosity >= 4: print("unsafe and skipped:",neighbour.tower)
    return safe_neighbours

def nodes_in_geometry(geometry):
    nodes = set()
    polygon = " ".join(map(lambda loc:str(loc[0])+" "+str(loc[1]),geometry))
    jsonfile = overpass("way(poly:%22"+polygon+"%22)[%22power%22=%22line%22];node(w)(poly:%22"+polygon+"%22);out;")
    for child in jsonfile['elements']:
        if child['type'] == "node":
            nodes.add(child['id'])
    return nodes

def get_towers_by_area(area):
    area_type = "addr:country" if area.isupper() else "name:en"
    query = "(area%5B%22"+area_type+"%22=%22"+area+"%22%5D;)-%3E.a;way(area.a)%5B%22power%22=%22line%22%5D;node(w);out;"
    jsonfile = overpass(query)
    towers = []
    for child in jsonfile['elements']:
        if child['type'] == "node":
            tower = pylon(child['id'],(float(child['lat']),float(child['lon'])))
            towers.append(tower)
    return towers
